from abc import ABC, abstractmethod
from typing import List, Dict, Type, Union, Optional
from pkg_resources import parse_version

from airflow.models import BaseOperator
from airflow.version import version as AIRFLOW_VERSION

from openlineage.client.facet import BaseFacet
from openlineage.common.dataset import Dataset

if parse_version(AIRFLOW_VERSION) >= parse_version("2.0.0"):
    # Corrects path of import for Airflow versions below 1.10.11
    from airflow.utils.log.logging_mixin import LoggingMixin
elif parse_version(AIRFLOW_VERSION) >= parse_version("1.10.11"):
    from airflow import LoggingMixin
else:
    # Corrects path of import for Airflow versions below 1.10.11
    from airflow.utils.log.logging_mixin import LoggingMixin


class StepMetadata:
    def __init__(
            self,
            name,
            location=None,
            inputs: List[Dataset] = None,
            outputs: List[Dataset] = None,
            context=None,
            run_facets: Dict[str, BaseFacet] = None
    ):
        # TODO: Define a common way across extractors to build the
        # job name for an operator
        self.name = name
        self.location = location
        self.inputs = inputs
        self.outputs = outputs
        self.context = context
        self.run_facets = run_facets

        if not inputs:
            self.inputs = []
        if not outputs:
            self.outputs = []
        if not context:
            self.context = {}
        if not run_facets:
            self.run_facets = {}

    def __repr__(self):
        return "name: {}\t inputs: {} \t outputs: {}".format(
            self.name,
            ','.join([str(i) for i in self.inputs]),
            ','.join([str(o) for o in self.outputs]))


class BaseExtractor(ABC, LoggingMixin):
    operator_class: Type[BaseOperator] = None
    operator: operator_class = None

    def __init__(self, operator):
        self.operator = operator
        self.patch()

    def patch(self):
        # Extractor should register extension methods or patches to operator here
        pass

    @classmethod
    def get_operator_class(cls):
        return cls.operator_class

    def validate(self):
        # TODO: maybe we should also enforce the module
        assert (self.operator_class is not None and
                self.operator.__class__ == self.operator_class)

    @abstractmethod
    def extract(self) -> Union[Optional[StepMetadata], List[StepMetadata]]:
        # In future releases, we'll want to deprecate returning a list of StepMetadata
        # and simply return a StepMetadata object. We currently return a list
        # for backwards compatibility.
        pass

    def extract_on_complete(self, task_instance) -> \
            Union[Optional[StepMetadata], List[StepMetadata]]:
        return self.extract()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import logging
from collections import defaultdict

import attr
from typing import Optional, Any, Dict, List

from openlineage.client.facet import BaseFacet

from openlineage.airflow.extractors.base import BaseExtractor, StepMetadata
from openlineage.airflow.facets import DataQualityDatasetFacet, ColumnMetric
from openlineage.airflow.utils import get_job_name
from openlineage.common.dataset import Source, Dataset
from openlineage.common.utils import get_from_nullable_chain


def wrap_callback(f):
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        result = f(self, *args, **kwargs)
        self._extractor.store_result(result)
        return result

    return wrapper


log = logging.getLogger(__file__)


# Great Expectations is optional dependency.
try:
    from great_expectations_provider.operators.great_expectations import GreatExpectationsOperator
    _has_great_expectations = True
    GreatExpectationsOperator.execute = wrap_callback(GreatExpectationsOperator.execute)
except Exception:
    # Create placeholder for GreatExpectationsOperator
    GreatExpectationsOperator = None
    log.warning('Did not find great_expectations_provider library or failed to import it')
    _has_great_expectations = False


@attr.s
class ExpectationsParserResult:
    """
    Internal class to represent actual expectation values, per table and optionally per column
    """
    facet_key: str = attr.ib()
    value: Any = attr.ib()
    column_id: Optional[str] = attr.ib(default=None)


@attr.s
class GreatExpectationsAssertion:
    expectationType: str = attr.ib()
    success: bool = attr.ib()
    column: Optional[str] = attr.ib(default=None)


@attr.s
class GreatExpectationsAssertionsDatasetFacet(BaseFacet):
    """
    This facet represents passed/failed status of asserted expectations on dataset
    """
    assertions: List[GreatExpectationsAssertion] = attr.ib()

    @staticmethod
    def _get_schema() -> str:
        return "https://github.com/OpenLineage/OpenLineage/tree/main/integration/airflow/openlineage/airflow/extractors/ge-assertions-dataset-facet.json"  # noqa


def set_dataset_info(operator: GreatExpectationsOperator, namespace: str, name: str):
    """
    Warning: you need to manually provide dataset namespace and name for this extractor
    in your dag. Due to technical limitation, without providing it, we can't find namespace
    and name of the dataset just by data extraction.

    Example:
    my_ge_task = GreatExpectationsOperator(
        task_id='my_task,
        expectation_suite_name='my_suite',
        batch_kwargs={
            'table': 'my_table',
            'datasource': 'my_datasource'
        },
        data_context_root_dir=ge_root_dir,
        dag=dag
    )

    set_dataset_info(my_ge_task, "postgres://postgres:5432", "database.public.my_table")
    """
    operator._expectations_namespace = namespace
    operator._expectations_name = name


class GreatExpectationsExtractorImpl(BaseExtractor):
    """
    Great Expectations extractor extracts validation data from CheckpointResult object and
    parses it via ExpectationsParsers. Results are used to prepare data quality facet.
    """
    operator_class = GreatExpectationsOperator

    def __init__(self, operator):
        super().__init__(operator)
        self.operator._extractor = self
        self.result = None

    def store_result(self, result):
        self.result = result

    def parse_result(self) -> Optional[Dataset]:
        try:
            run_result = self.result['run_results']
            # Great expectations generate long, impenetrable name here
            validation_result = next(iter(run_result.values()))['validation_result']

            data_quality_facet = self.parse_data_quality_facet(validation_result)
            if not data_quality_facet:
                return None

            assertions_facet = self.parse_assertions(validation_result)
            if not assertions_facet:
                return None

            batch_kwargs = get_from_nullable_chain(
                validation_result,
                ['meta', 'batch_kwargs']
            )

            scheme, authority, namespace, name = None, None, None, None

            # To match dataset name we need canonical datasource name
            if hasattr(self.operator, '_expectations_namespace'):
                namespace = self.operator._expectations_namespace
                name = self.operator._expectations_name
                if '://' in namespace:
                    scheme, authority = namespace.split('://')
                elif ':' in namespace:
                    scheme, authority = namespace.split(':')
                else:
                    scheme = namespace
            else:
                name = batch_kwargs.get('datasource', None)
                path = batch_kwargs.get('path', None)
                namespace = f'{name}:{path}'
            return Dataset(
                source=Source(
                    scheme=scheme,
                    authority=authority,
                    name=namespace if not scheme and authority else None
                ),
                name=name,
                custom_facets={
                    'dataQuality': data_quality_facet,
                    'greatExpectations_assertions': assertions_facet
                },
                input_facets={
                    'dataQualityMetrics': data_quality_facet.to_openlineage()
                }
            )

        except ValueError:
            log.exception("Exception while retrieving great expectations dataset")
        return None

    def parse_data_quality_facet(self, validation_result: Dict) \
            -> Optional[DataQualityDatasetFacet]:
        facet_data = {
            "columnMetrics": defaultdict(dict)
        }

        # try to get to actual expectations results
        try:
            expectations_results = validation_result['results']
            for expectation in expectations_results:
                for parser in _EXPECTATIONS_PARSERS:

                    # accept possible duplication, should have no difference in results
                    if parser.can_accept(expectation):
                        result = parser.parse_expectation_result(expectation)
                        facet_data[result.facet_key] = result.value
                for parser in _COLUMN_EXPECTATIONS_PARSER:
                    if parser.can_accept(expectation):
                        result = parser.parse_expectation_result(expectation)
                        facet_data['columnMetrics'][result.column_id][result.facet_key] \
                            = result.value

            for key in facet_data['columnMetrics'].keys():
                facet_data['columnMetrics'][key] = ColumnMetric(**facet_data['columnMetrics'][key])
            return DataQualityDatasetFacet(**facet_data)
        except ValueError:
            log.exception(
                "Great Expectations's CheckpointResult object does not have expected key"
            )
        return None

    def parse_assertions(self, validation_result: Dict) -> \
            Optional[GreatExpectationsAssertionsDatasetFacet]:
        assertions = []

        try:
            expectations_results = validation_result['results']
            for expectation in expectations_results:
                assertions.append(GreatExpectationsAssertion(
                    expectationType=expectation['expectation_config']['expectation_type'],
                    success=expectation['success'],
                    column=expectation['expectation_config']['kwargs'].get('column', None)
                ))

            return GreatExpectationsAssertionsDatasetFacet(assertions)
        except ValueError:
            log.exception(
                "Great Expectations's CheckpointResult object does not have expected key"
            )
        return None

    def extract(self) -> Optional[StepMetadata]:
        return None

    def extract_on_complete(self, task_instance) -> Optional[StepMetadata]:
        if self.result:
            dataset = self.parse_result()
            if not dataset:
                return None
            return StepMetadata(
                name=get_job_name(task=self.operator),
                inputs=[dataset]
            )
        return None


class ExpectationsParser:
    """
    Base expectation parser. Dispatches parser looking at expectation type.
    Implementations should extract result from result dictionary.
    """
    expectation_key: str = ''
    facet_key: str = ''

    @classmethod
    def can_accept(cls, expectation_results: dict) -> bool:
        expectation_type = get_from_nullable_chain(
            expectation_results,
            ['expectation_config', 'expectation_type']
        )
        return expectation_type and expectation_type == cls.expectation_key

    @staticmethod
    def parse_expectation_result(expectation_result: dict) -> ExpectationsParserResult:
        raise NotImplementedError("")


class BetweenRowCountExpectationsParser(ExpectationsParser):
    expectation_key = 'expect_table_row_count_to_be_between'
    facet_key = 'rowCount'

    @classmethod
    def parse_expectation_result(cls, expectation_result: dict) -> ExpectationsParserResult:
        return ExpectationsParserResult(
            cls.facet_key,
            get_from_nullable_chain(
                expectation_result,
                ['result', 'observed_value']
            )
        )


class EqualRowCountExpectationsParser(BetweenRowCountExpectationsParser):
    expectation_key = 'expect_table_row_count_to_equal'


class FileSizeExpectationsParser(ExpectationsParser):
    expectation_key = 'expect_file_size_to_be_between'

    @staticmethod
    def parse_expectation_result(expectation_result: dict) -> ExpectationsParserResult:
        pass  # TODO: file asset validation


class ColumnExpectationsParser(ExpectationsParser):
    """
    Extractor for column-based expectations. Looks at column name in addition to expectation type
    """
    column = ''

    @classmethod
    def can_accept(cls, expectation_results: dict) -> bool:
        expectation_type = get_from_nullable_chain(
            expectation_results,
            ['expectation_config', 'expectation_type']
        )
        extracted_column = get_from_nullable_chain(
            expectation_results,
            ['expectation_config', 'kwargs', 'column']
        )
        return expectation_type and extracted_column \
            and expectation_type == cls.expectation_key


class ValuesNotNullColumnExpectationParser(ColumnExpectationsParser):
    expectation_key = 'expect_column_values_to_not_be_null'

    @staticmethod
    def parse_expectation_result(expectation_result: dict) -> Any:
        return ExpectationsParserResult(
            'nullCount',
            get_from_nullable_chain(expectation_result, ['result', 'unexpected_count']),
            get_from_nullable_chain(
                expectation_result,
                ['expectation_config', 'kwargs', 'column']
            )
        )


class ValuesDistinctExpectationParser(ColumnExpectationsParser):
    expectation_key = 'expect_column_unique_value_count_to_be_between'

    @staticmethod
    def parse_expectation_result(expectation_result: dict) -> ExpectationsParserResult:
        return ExpectationsParserResult(
            'distinctCount',
            get_from_nullable_chain(expectation_result, ['result', 'observed_value']),
            get_from_nullable_chain(
                expectation_result,
                ['expectation_config', 'kwargs', 'column']
            )
        )


class ValuesSumExpectationParser(ColumnExpectationsParser):
    expectation_key = 'expect_column_sum_to_be_between'

    @staticmethod
    def parse_expectation_result(expectation_result: dict) -> ExpectationsParserResult:
        sum = get_from_nullable_chain(expectation_result, ['result', 'observed_value'])
        return ExpectationsParserResult(
            'sum',
            sum,
            get_from_nullable_chain(
                expectation_result,
                ['expectation_config', 'kwargs', 'column']
            )
        )


class ValuesCountExpectationParser(ColumnExpectationsParser):
    expectation_key = 'expect_column_sum_to_be_between'

    @staticmethod
    def parse_expectation_result(expectation_result: dict) -> ExpectationsParserResult:
        count = get_from_nullable_chain(expectation_result, ['result', 'element_count'])
        return ExpectationsParserResult(
            'count',
            count,
            get_from_nullable_chain(
                expectation_result,
                ['expectation_config', 'kwargs', 'column']
            )
        )


class ValuesMaxExpectationParser(ColumnExpectationsParser):
    expectation_key = 'expect_column_max_to_be_between'

    @staticmethod
    def parse_expectation_result(expectation_result: dict) -> ExpectationsParserResult:
        return ExpectationsParserResult(
            'max',
            get_from_nullable_chain(expectation_result, ['result', 'observed_value']),
            get_from_nullable_chain(
                expectation_result,
                ['expectation_config', 'kwargs', 'column']
            )
        )


class ValuesMinExpectationParser(ColumnExpectationsParser):
    expectation_key = 'expect_column_min_to_be_between'

    @staticmethod
    def parse_expectation_result(expectation_result: dict) -> ExpectationsParserResult:
        return ExpectationsParserResult(
            'min',
            get_from_nullable_chain(expectation_result, ['result', 'observed_value']),
            get_from_nullable_chain(
                expectation_result,
                ['expectation_config', 'kwargs', 'column']
            )
        )


class ValuesQuantileExpectationParser(ColumnExpectationsParser):
    expectation_key = 'expect_column_quantile_values_to_be_between'

    @staticmethod
    def quantile_to_map(observations):
        return {
            str(k): v for k, v in zip(observations['quantiles'], observations['values'])
        }

    @classmethod
    def parse_expectation_result(cls, expectation_result: dict) -> ExpectationsParserResult:
        observed_values = get_from_nullable_chain(expectation_result, ['result', 'observed_value'])
        return ExpectationsParserResult(
            'quantiles',
            cls.quantile_to_map(
                observed_values
            ) if observed_values else None,
            get_from_nullable_chain(
                expectation_result,
                ['expectation_config', 'kwargs', 'column']
            )
        )


_EXPECTATIONS_PARSERS = [
    BetweenRowCountExpectationsParser,
    EqualRowCountExpectationsParser,
    # FileSizeExpectationsParser,
]

_COLUMN_EXPECTATIONS_PARSER = [
    ValuesNotNullColumnExpectationParser,
    ValuesDistinctExpectationParser,
    ValuesMinExpectationParser,
    ValuesMaxExpectationParser,
    ValuesQuantileExpectationParser,
    ValuesSumExpectationParser,
    ValuesCountExpectationParser
]

if _has_great_expectations:
    GreatExpectationsExtractor = GreatExpectationsExtractorImpl
else:
    class GreatExpectationsExtractor:
        def __init__(self):
            raise RuntimeError('Great Expectations provider not found')



import attr
from airflow.version import version as AIRFLOW_VERSION
from openlineage.airflow import __version__ as OPENLINEAGE_AIRFLOW_VERSION
from openlineage.client.facet import BaseFacet, DataQualityMetricsInputDatasetFacet
from typing import Optional, Dict


@attr.s
class AirflowVersionRunFacet(BaseFacet):
    operator: str = attr.ib()
    taskInfo: str = attr.ib()
    airflowVersion: str = attr.ib()
    openlineageAirflowVersion: str = attr.ib()

    @classmethod
    def from_task(cls, task):
        return cls(
            f'{task.__class__.__module__}.{task.__class__.__name__}',
            str(task.__dict__),
            AIRFLOW_VERSION,
            OPENLINEAGE_AIRFLOW_VERSION
        )


@attr.s
class AirflowRunArgsRunFacet(BaseFacet):
    externalTrigger: bool = attr.ib(default=False)


@attr.s
class ColumnMetric:
    nullCount: Optional[int] = attr.ib(default=None)
    distinctCount: Optional[int] = attr.ib(default=None)
    sum: Optional[float] = attr.ib(default=None)
    count: Optional[int] = attr.ib(default=None)
    min: Optional[float] = attr.ib(default=None)
    max: Optional[float] = attr.ib(default=None)
    quantiles: Optional[Dict[str, float]] = attr.ib(default=None)


@attr.s
class DataQualityDatasetFacet(BaseFacet):
    rowCount: Optional[int] = attr.ib(default=None)
    bytes: Optional[int] = attr.ib(default=None)
    columnMetrics: Dict[str, ColumnMetric] = attr.ib(factory=dict)

    def to_openlineage(self) -> DataQualityMetricsInputDatasetFacet:
        return DataQualityMetricsInputDatasetFacet(
            rowCount=self.rowCount,
            bytes=self.bytes,
            columnMetrics=self.columnMetrics
        )

    @staticmethod
    def _get_schema() -> str:
        return "https://github.com/OpenLineage/OpenLineage/tree/main/integration/airflow/openlineage/airflow/data-quality-dataset-facet.json"  # noqa



from typing import Type, Optional

from openlineage.airflow.extractors.base import BaseExtractor
from openlineage.airflow.extractors.bigquery_extractor import BigQueryExtractor
from openlineage.airflow.extractors.great_expectations_extractor import GreatExpectationsExtractor
from openlineage.airflow.extractors.postgres_extractor import PostgresExtractor
from openlineage.airflow.extractors.snowflake_extractor import SnowflakeExtractor


_extractors = [
    PostgresExtractor,
    BigQueryExtractor,
    GreatExpectationsExtractor,
    SnowflakeExtractor
]

_patchers = [
    GreatExpectationsExtractor
]


class Extractors:
    """
    This exposes implemented extractors, while hiding ones that require additional, unmet
    dependency. Patchers are a category of extractor that needs to hook up to operator's
    internals during DAG creation.
    """
    def __init__(self):
        # Do not expose extractors relying on external dependencies that are not installed
        self.extractors = {
            extractor.operator_class: extractor
            for extractor
            in _extractors
            if getattr(extractor, 'operator_class', None) is not None
        }

        self.patchers = {
            extractor.operator_class: extractor
            for extractor
            in _patchers
            if getattr(extractor, 'operator_class', None) is not None
        }

    def get_extractor_class(self, clazz: Type) -> Optional[Type[BaseExtractor]]:
        if clazz in self.extractors:
            return self.extractors[clazz]
        return None

    def get_patcher_class(self, clazz: Type) -> Optional[Type[BaseExtractor]]:
        if clazz in self.patchers:
            return self.patchers[clazz]
        return None



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import subprocess
from uuid import uuid4
from urllib.parse import urlparse, urlunparse

import airflow
from airflow.models import Connection
from airflow.utils.db import provide_session
from openlineage.airflow.facets import AirflowVersionRunFacet, AirflowRunArgsRunFacet

try:
    # Import from pendulum 1.x version
    from pendulum import Pendulum, from_timestamp
except ImportError:
    # Import for Pendulum 2.x version
    from pendulum import DateTime as Pendulum, from_timestamp

log = logging.getLogger(__name__)
_NOMINAL_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


class JobIdMapping:
    # job_name here is OL job name - aka combination of dag_id and task_id

    @staticmethod
    def set(job_name: str, dag_run_id: str, task_run_id: str):
        airflow.models.Variable.set(
            JobIdMapping.make_key(job_name, dag_run_id),
            json.dumps(task_run_id)
        )

    @staticmethod
    def pop(job_name, dag_run_id, session):
        return JobIdMapping.get(job_name, dag_run_id, session, delete=True)

    @staticmethod
    def get(job_name, dag_run_id, session, delete=False):
        key = JobIdMapping.make_key(job_name, dag_run_id)
        if session:
            q = session.query(airflow.models.Variable).filter(
                airflow.models.Variable.key == key)
            if not q.first():
                return None
            else:
                val = q.first().val
                if delete:
                    q.delete(synchronize_session=False)
                if val:
                    return json.loads(val)
                return None

    @staticmethod
    def make_key(job_name, run_id):
        return "openlineage_id_mapping-{}-{}".format(job_name, run_id)


def url_to_https(url) -> str:
    # Ensure URL exists
    if not url:
        return None

    base_url = None
    if url.startswith('git@'):
        part = url.split('git@')[1:2]
        if part:
            base_url = f'https://{part[0].replace(":", "/", 1)}'
    elif url.startswith('https://'):
        base_url = url

    if not base_url:
        raise ValueError(f"Unable to extract location from: {url}")

    if base_url.endswith('.git'):
        base_url = base_url[:-4]
    return base_url


def get_location(file_path) -> str:
    # Ensure file path exists
    if not file_path:
        return None

    # move to the file directory
    abs_path = os.path.abspath(file_path)
    file_name = os.path.basename(file_path)
    cwd = os.path.dirname(abs_path)

    # get the repo url
    repo_url = execute_git(cwd, ['config', '--get', 'remote.origin.url'])

    # get the repo relative path
    repo_relative_path = execute_git(cwd, ['rev-parse', '--show-prefix'])

    # get the commitId for the particular file
    commit_id = execute_git(cwd, ['rev-list', 'HEAD', '-1', '--', file_name])

    # build the URL
    base_url = url_to_https(repo_url)
    if not base_url:
        return None

    return f'{base_url}/blob/{commit_id}/{repo_relative_path}{file_name}'


def execute_git(cwd, params):
    p = subprocess.Popen(['git'] + params,
                         cwd=cwd, stdout=subprocess.PIPE, stderr=None)
    p.wait(timeout=0.5)
    out, err = p.communicate()
    return out.decode('utf8').strip()


def get_connection_uri(conn: Connection):
    """
    Return the connection URI for the given ID. We first attempt to lookup
    the connection URI via AIRFLOW_CONN_<conn_id>, else fallback on querying
    the Airflow's connection table.
    """

    conn_uri = conn.get_uri()
    parsed = urlparse(conn_uri)

    # Remove username and password
    parsed = parsed._replace(netloc=f'{parsed.hostname}:{parsed.port}')
    return urlunparse(parsed)


def get_normalized_postgres_connection_uri(conn: Connection):
    """
    URIs starting with postgresql:// and postgres:// are both valid
    PostgreSQL connection strings. This function normalizes it to
    postgres:// as canonical name according to OpenLineage spec.
    """
    uri = get_connection_uri(conn)
    if uri.startswith('postgresql'):
        uri = uri.replace('postgresql', 'postgres', 1)
    return uri


@provide_session
def get_connection(conn_id, session=None) -> Connection:
    # TODO: We may want to throw an exception if the connection
    # does not exist (ex: AirflowConnectionException). The connection
    # URI is required when collecting metadata for a data source.
    conn_uri = os.environ.get('AIRFLOW_CONN_' + conn_id.upper())
    if conn_uri:
        conn = Connection()
        conn.parse_from_uri(conn_uri)
        return conn

    return (session
            .query(Connection)
            .filter(Connection.conn_id == conn_id)
            .first())


def get_job_name(task):
    return f'{task.dag_id}.{task.task_id}'


def get_custom_facets(task, is_external_trigger: bool):
    return {
        "airflow_runArgs": AirflowRunArgsRunFacet(is_external_trigger),
        "airflow_version": AirflowVersionRunFacet.from_task(task)
    }


def new_lineage_run_id(dag_run_id: str, task_id: str) -> str:
    return str(uuid4())


class DagUtils:

    def get_execution_date(**kwargs):
        return kwargs.get('execution_date')

    @staticmethod
    def get_start_time(execution_date=None):
        if execution_date:
            return DagUtils.to_iso_8601(execution_date)
        else:
            return None

    @staticmethod
    def get_end_time(execution_date, default):
        if execution_date:
            end_time = default
        else:
            end_time = None

        if end_time:
            end_time = DagUtils.to_iso_8601(end_time)
        return end_time

    @staticmethod
    def to_iso_8601(dt):
        if not dt:
            return None
        if isinstance(dt, int):
            dt = from_timestamp(dt/1000.0)

        if isinstance(dt, Pendulum):
            return dt.format(_NOMINAL_TIME_FORMAT)
        else:
            return dt.strftime(_NOMINAL_TIME_FORMAT)
